class Plugboard:
    """Plugboard (Steckerbrett in German) of an Enigma."""
    def __init__(self, parent=None):
        self.map_dict = {}
        self.parent = parent

    def check_parent(self):
        if self.parent is None:
            raise RuntimeError('parent of the Plugboard is None')

    def plug(self, letter_1: str, letter_2: str):
        if letter_1 == letter_2:
            raise AttributeError('Two letters can\'t be equal')
        for letter in (letter_1, letter_2):
            if letter not in self.parent.entry_plate.map_source:
                raise AttributeError(f'"{letter}" is not in map_source')
            if letter in self.map_dict and self.map_dict[letter]:
                raise RuntimeError(f'"{letter}" have been set to be connected with another letter')
        self.map_dict[letter_1] = letter_2
        self.map_dict[letter_2] = letter_1

    def unplug(self, letter: str):
        if letter not in self.parent.entry_plate.map_source:
            raise AttributeError(f'"{letter}" is not in map_source')
        if letter not in self.map_dict:
            raise RuntimeError(f'"{letter}" have not been set')
        another_letter = self.map_dict[letter]
        del self.map_dict[letter]
        del self.map_dict[another_letter]



from ..part.plate import Rotor


def rotor_i():
    """Rotor I."""
    return Rotor('EKMFLGDQVZNTOWYHXUSPAIBRCJ', name='Rotor I', turnover='Q')


def rotor_ii():
    """Rotor II."""
    return Rotor('AJDKSIRUXBLHWTMCQGZNPYFVOE', name='Rotor II', turnover='E')


def rotor_iii():
    """Rotor III."""
    return Rotor('BDFHJLCPRTXVZNYEIWGAKMUSQO', name='Rotor III', turnover='V')


def rotor_iv():
    """Rotor IV."""
    return Rotor('ESOVPZJAYQUIRHXLNFTGKDCMWB', name='Rotor IV', turnover='J')


def rotor_v():
    """Rotor V."""
    return Rotor('VZBRGITYUPSDNHLXAWMJQOFECK', name='Rotor V', turnover='Z')


def rotor_vi():
    """Rotor VI."""
    return Rotor('JPGVOUMFYQBENHZRDKASXLICTW', name='Rotor VI', turnover=['Z', 'M'])


def rotor_vii():
    """Rotor VII."""
    return Rotor('NZJHGRCXMYSWBOUFAIVLPEKQDT', name='Rotor VII', turnover=['Z', 'M'])


def rotor_viii():
    """Rotor VIII."""
    return Rotor('FKQHTLXOCBJSPDZRAMEWNIUYGV', name='Rotor VIII', turnover=['Z', 'M'])


def rotor_beta():
    """Rotor Beta."""
    return Rotor('LEYJVCNIXWPBQMDRTAKZGFUHOS', name='Rotor Beta')


def rotor_gamma():
    """Rotor Gamma."""
    return Rotor('FSOKANUERHMBTIYCWLQPZXVGJD', name='Rotor Gamma')



from .plate import Plate, _ALPHABET


class Rotor(Plate):
    """Rotor (Walzen in German) of an Enigma."""

    def __turnover_attr_exception(
            self, turnover: int | str | list[int | str] | tuple[int | str], map_source: str
    ):
        """Raise AttributeError if the turnover value is invalid."""
        raise AttributeError(
            f'turnover "{turnover}" is invalid. '
            f'A letter string in map_source or an int between 0 and {len(map_source) - 1} is required.'
        )

    def turnover_single_check_and_standardize(self, turnover_single: int | str, map_source: str) -> int:
        """Check single turnover value and standardize it.

        :return: int: index of the letter.
        :raise AttributeError: If the value is invalid.
        """
        if isinstance(turnover_single, str):
            if turnover_single in map_source:
                return map_source.index(turnover_single)
        elif isinstance(turnover_single, int):
            if 0 <= turnover_single < len(map_source):
                return turnover_single
        self.__turnover_attr_exception(turnover_single, map_source)

    def __init__(
            self, map_table: str = _ALPHABET, init_position: int | str = 0, auto_rotatable: bool = True,
            right_plate: Plate = None, left_plate: Plate = None, rotate_up: bool = False, name: str = None,
            map_source: str = _ALPHABET, turnover: int | str | list[int | str] | tuple[int | str] = None
    ):
        """Defining a plate.

        :param map_table: String for map table.
        :param init_position: Initial position.
        :param auto_rotatable: If the plate is auto rotatable.
            If it's True, the plate will be rotatable if other plate let it rotate.
            Or it's False, but you can also let it rotate by setting its position.
        :param right_plate: The plate on the right of this plate.
        :param left_plate: The plate on the left of this plate.
        :param rotate_up: Rotate direction of this plate.
            If it's True, it will rotate from small to large by index.
                (If upper is A, lower is B, the plate will rotate from lower to upper when change from A to B.)
            Or it's False (Default).
                (If upper is B, lower is A, the plate will rotate from upper to lower when change from A to B.)
        :param name: The name of the plate.
        :param map_source: String for map source. the circuits are:
            item in map_source -> item in map_table
        :param turnover: Define turnover position.
            For example, if it's set to 'T',
                it will let the left rotor rotate a step when this rotor rotates from T to U (in alphabet).
        :raise AttributeError: If map_table or init_position is invalid.
        """
        super().__init__(map_table, init_position, auto_rotatable, right_plate, left_plate, rotate_up, name,
                         map_source=map_source)
        if turnover is None:
            turnover = map_source[-1]
        if isinstance(turnover, list) or isinstance(turnover, tuple):
            turnover_list = []
            for turnover_item in turnover:
                turnover_list.append(self.turnover_single_check_and_standardize(turnover_item, map_source))
        else:
            turnover_list = [self.turnover_single_check_and_standardize(turnover, map_source)]
        self.turnover = turnover_list

    def forward(self) -> int:
        """Forward the plate.

        :return: Position of the plate.
        """
        original_position = self.position
        if original_position == len(self.map_source) - 1:
            self.position = 0
        else:
            self.position += 1
        if original_position in self.turnover:
            if self.left_plate:
                if self.left_plate.auto_rotatable:
                    self.left_plate.forward()
        return self.position



from .plate import Plate, _ALPHABET


class Reflector(Plate):
    """Reflector (Umkehrwalze in German) of an Enigma."""

    def __init__(
            self, map_table: str = _ALPHABET, init_position: str | int = 0, right_plate: Plate = None,
            name: str = None, map_source: str = _ALPHABET
    ):
        """Defining a reflector.

        :param map_table: String for map table.
        :param init_position: Initial position.
        :param right_plate: The plate on the right of this plate.
        :param name: The name of the plate.
        :param map_source: String for map source. the circuits are:
            item in map_source -> item in map_table
        :raise AttributeError: If map_table or init_position is invalid.
        """
        super().__init__(map_table, init_position, False, right_plate, None, name=name, map_source=map_source)



from .plate import Plate, _ALPHABET


class EntryPlate(Plate):
    """Entry Plate of an Enigma."""

    def __init__(
            self, map_table: str = _ALPHABET, init_position: str | int = 0, left_plate: Plate = None,
            name: str = 'Entry Plate'
    ):
        super().__init__(map_table, init_position, False, None, left_plate, name=name, map_source=map_table)



from .part import Plugboard
from .part.plate import Rotor, EntryPlate, Reflector


class Enigma:
    """An Enigma machine."""

    def __init__(
            self, rotor_list: list[Rotor] | tuple[Rotor], reflector: Reflector, rotate_up: bool = False,
            rotate_after_type: bool = False
    ):
        """Define the Enigma.

        :param rotor_list: The list of the rotors from right to left. Tuple with rotors is also supported.
        :param reflector: The reflector of the Enigma.
        :param rotate_up: Rotate direction of the rotors.
            If it's True, it will rotate from small to large by index.
                (If upper is A, lower is B, the plate will rotate from lower to upper when change from A to B.)
            Or it's False (Default).
                (If upper is B, lower is A, the plate will rotate from upper to lower when change from A to B.)
        :param rotate_after_type: Set if the rotors rotate after type. Defaults to False.
        :raise AttributeError: If the map_source of any rotor is not equal to that in reflector.
        """
        self.rotors = list(rotor_list)
        self.reflector = reflector
        self.reflector.right_plate = self.rotors[-1]
        self.entry_plate = EntryPlate(map_table=self.reflector.map_source, left_plate=self.rotors[0])
        for rotor_index, rotor in enumerate(self.rotors):
            if rotor.map_source != self.reflector.map_source:
                raise AttributeError(f'map_source of rotor[{rotor_index}] is not equal to that in reflector')
            if rotor_index < len(self.rotors) - 1:
                rotor.left_plate = self.rotors[rotor_index + 1]
            else:
                rotor.left_plate = self.reflector
            if rotor_index > 0:
                rotor.right_plate = self.rotors[rotor_index - 1]
            else:
                rotor.right_plate = self.entry_plate
        for rotor in self.rotors:
            rotor.rotate_up = rotate_up
        self.rotate_after_type = rotate_after_type
        self.plugboard = Plugboard(parent=self)

    def input(self, string: str) -> str:
        """Input string to Enigma.

        :param string: String.
        :return: String of encrypted result.
        """
        new_string_list = []
        for letter in string:
            if not self.rotate_after_type:
                self.rotors[0].forward()
            if letter in self.plugboard.map_dict:
                letter = self.plugboard.map_dict[letter]
            for rotor in self.rotors:
                letter = rotor.encrypt(letter, 'right')
            letter = self.reflector.encrypt(letter, 'right')
            for rotor in reversed(self.rotors):
                letter = rotor.encrypt(letter, 'left')
            letter = self.entry_plate.encrypt(letter, 'left')
            if letter in self.plugboard.map_dict:
                letter = self.plugboard.map_dict[letter]
            new_string_list.append(letter)
            if self.rotate_after_type:
                self.rotors[0].forward()
        return ''.join(new_string_list)

    def set_position(self, *position_list):
        """Set positions of the rotors.

        :param position_list: the position of rotors from right to left.
        :raise AttributeError: If position list length is not equal to rotors number.
        """
        if not position_list:
            position_list = (0, ) * len(self.rotors)
        elif len(position_list) != len(self.rotors):
            raise AttributeError('Position list length is not equal to rotors number')
        for rotor, position in zip(self.rotors, position_list):
            rotor.set_position(position)



from ..part.plate import Reflector


def reflector_b():
    """Reflector B."""
    return Reflector('YRUHQSLDPXNGOKMIEBFZCWVJAT', name='Reflector B')


def reflector_c():
    """Reflector C."""
    return Reflector('FVPJIAOYEDRZXWGCTKUQSBNMHL', name='Reflector C')


def reflector_b_thin():
    """Reflector B Thin."""
    return Reflector('ENKQAUYWJICOPBLMDXZVFTHRGS', name='Reflector B Thin')


def reflector_c_thin():
    """Reflector C Thin."""
    return Reflector('RDOBJNTKVEHMLFCWZAXGYIPSUQ', name='Reflector C Thin')



from typing import Optional

_ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'


class Plate:
    """Plate of an Enigma, including Entry Plate, Rotors (Walzen in German) and Reflector (Umkehrwalze in German)."""

    @staticmethod
    def _map_table_check(map_table: str, map_source: str = _ALPHABET):
        """Raise AttributeError if map table is invalid."""
        if len(set(map_source)) != len(map_source):
            raise AttributeError(f'map_source "{map_source}" is invalid. It must be str with different letters.')
        if len(set(map_table)) != len(map_source) or len(map_table) != len(map_source):
            raise AttributeError(
                f'map_table "{map_table}" is invalid. '
                f'It must be str with different letters, and every letter is in map_source "{map_source}".'
            )
        for letter in map_source:
            if letter not in map_table:
                raise AttributeError(f'"{letter}" not in map table.')

    def __init_init_position_attr_exception(self, init_position: int | str, map_source: str):
        """Raise AttributeError if init_position is invalid."""
        raise AttributeError(
            f'init_position "{init_position}" is invalid. '
            f'A letter string in map_source or an int between 0 and {len(map_source) - 1} is required.'
        )

    def __init__(
            self, map_table: str = _ALPHABET, init_position: str | int = 0, auto_rotatable: bool = False,
            right_plate=None, left_plate=None, rotate_up: bool = False, name: str = None, map_source: str = _ALPHABET
    ):
        """Defining a plate.

        :param map_table: String for map table.
        :param init_position: Initial position.
        :param auto_rotatable: If the plate is auto rotatable.
            If it's True, the plate will be rotatable if other plate let it rotate.
            Or it's False, but you can also let it rotate by setting its position.
        :param right_plate: The plate on the right of this plate.
        :param left_plate: The plate on the left of this plate.
        :param rotate_up: Rotate direction of this plate.
            If it's True, it will rotate from small to large by index.
                (If upper is A, lower is B, the plate will rotate from lower to upper when change from A to B.)
            Or it's False (Default).
                (If upper is B, lower is A, the plate will rotate from upper to lower when change from A to B.)
        :param name: The name of the plate.
        :param map_source: String for map source. the circuits are:
            item in map_source -> item in map_table
        :raise AttributeError: If map_table or init_position is invalid.
        """
        self._map_table_check(map_table, map_source)
        dict_table = {}
        for input_letter_index, input_letter in enumerate(map_source):
            dict_table[input_letter] = map_table[input_letter_index]
        self.dict_table = dict_table
        if type(init_position) == str:
            if len(init_position) != 1 or init_position not in map_source:
                self.__init_init_position_attr_exception(init_position, map_source)
            self.position = map_source.index(init_position)
        elif type(init_position) == int:
            if 0 <= init_position < len(map_table):
                self.position = init_position
            else:
                self.__init_init_position_attr_exception(init_position, map_source)
        else:
            self.__init_init_position_attr_exception(init_position, map_source)
        self.auto_rotatable = auto_rotatable
        self.right_plate = right_plate
        self.left_plate = left_plate
        self.rotate_up = rotate_up
        self.name = name
        self.map_source = map_source
        self.turnover = []

    @property
    def current_state(self) -> dict:
        """Return a dict with 'left' and 'right' keys, which show the current state."""
        right = list(self.dict_table.keys())
        left = list(self.dict_table.values())
        transform_position = (-2 * self.rotate_up + 1) * self.position
        current_right = right[transform_position:] + right[:transform_position]
        current_left = left[transform_position:] + left[:transform_position]
        return {'left': current_left, 'right': current_right}

    def transform_letter(self, letter: str, from_plate) -> str:
        """Transform letter from previous process to the letter on the plate.

        :param letter: Letter from the previous process.
        :param from_plate: 'left' or 'right'
        :return: transformed letter.
        """
        letter_index = from_plate.current_state['right'].index(letter)
        return self.current_state['right'][letter_index]

    def encrypt(self, letter: str, letter_from: str = 'right') -> str:
        """Return the output letter through the plate.

        :param letter: A letter.
        :param letter_from: 'left' or 'right'.
        :return: A letter.
        :raise AttributeError: Input invalid
        """
        if type(letter) != str or len(letter) != 1 or letter not in self.map_source:
            raise AttributeError(f'Input "{letter}" is invalid. A letter string in map_source is required.')
        if letter_from == 'left':
            from_plate = self.left_plate
        else:
            from_plate = self.right_plate
        if from_plate:
            letter = self.transform_letter(letter, from_plate)
        if letter_from == 'left':
            key_index = list(self.dict_table.values()).index(letter)
            left_letter = list(self.dict_table.keys())[key_index]
            return left_letter
        else:
            return self.dict_table[letter]

    def set_position(self, position: int | str):
        """Set position of the plate.

        :param position: Index or letter.
        :raise AttributeError: If position is invalid.
        """
        position_int: Optional[int] = None
        if isinstance(position, str):
            if position in self.map_source:
                position_int = self.map_source.index(position)
        elif isinstance(position, int):
            if 0 <= position < len(self.map_source):
                position_int = position
        if position_int is None:
            raise AttributeError(
                f'position "{position}" is invalid. '
                f'A letter string in map_source or an int between 0 and {len(self.map_source) - 1} is required.'
            )
        self.position = position_int
